"""

Oanda client


"""

from easyoanda import easyoanda
from loguru import logger

from ._client import CexClient


class OandaHandler(CexClient):
    """
    library: https://pypi.org/project/easyoanda/

    Args:
        None

    Returns:
        None

    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initialize the client

        """
        super().__init__(**kwargs)
        self.client = easyoanda.start_session(
            sessionType="paper",
            accountID=self.user_id,
            token=self.api_key,
        )

    async def get_quote(self, instrument):
        """
        Return a quote for a instrument


        Args:
            cex
            instrument

        Returns:
            quote
        """
        if quotes := self.client.pricing.start_stream(instrument):
            quote = quotes[0]
            return quote["bids"][0]["price"]

    async def get_account_balance(self):
        """
        return account balance

        Args:
            None

        Returns:
            balance

        """
        self.accounts_data = self.client.account.get_summary()
        return self.accounts_data["account"]["balance"]

    async def get_account_position(self):
        """
        Return account position.
        of a given exchange

        Args:
            None

        Returns:
            position

        """

        position_data = self.client.positions.openPositions()
        return position_data["positions"]

    async def get_trading_asset_balance(self):
        """ """
        self.accounts_data = self.client.account.get_summary()
        return self.accounts_data["account"]["balance"]

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """
        action = order_params.get("action")
        instrument = await self.replace_instrument(order_params.get("instrument"))
        quantity = order_params.get("quantity", self.trading_risk_amount)
        if not action or not instrument:
            raise ValueError("Missing required parameters: action or instrument")
        logger.debug(f"quantity {quantity}")
        amount = await self.get_order_amount(
            quantity=quantity,
            instrument=instrument,
            is_percentage=self.trading_risk_percentage,
        )
        if not amount:
            raise ValueError("Failed to calculate order amount")
        logger.debug(f"amount {amount}")
        pre_order_checks = await self.pre_order_checks(order_params)
        if not pre_order_checks:
            raise ValueError("Pre-order checks failed")

        if action == "SELL":
            amount = -amount

        marketOrder = easyoanda.MarketOrder()
        marketOrder.set(instrument=instrument, units=amount)
        self.client.orders.place_order(marketOrder)



"""

CCXT client


"""

import ccxt
from loguru import logger

from ._client import CexClient


class CcxtHandler(CexClient):
    """
    Library: https://github.com/ccxt/ccxt

    Args:
        None

    Returns:
        None

    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initialize the client

        """
        super().__init__(**kwargs)
        if self.name is None:
            return
        client = getattr(ccxt, self.name)
        self.balance_limit = kwargs.get("balance_limit", True)
        self.balance_limit_value = kwargs.get("balance_limit_value", 5)
        self.client = client(
            {
                "apiKey": self.api_key,
                "secret": self.secret,
                "password": self.password,
                "enableRateLimit": True,
                "options": {
                    "defaultType": self.defaulttype,
                },
            }
        )
        if self.testmode:
            self.client.set_sandbox_mode("enabled")
        self.account_number = self.client.uid
        self.name = self.client.id

    async def get_quote(self, instrument):
        """
        Asynchronously fetches a ask/offer quote
        for the specified instrument.

        :param instrument: The instrument for which the quote is to be fetched.
        :return: The fetched quote.
        """
        try:
            instrument = await self.replace_instrument(instrument)
            quote = self.client.fetch_ticker(instrument)["ask"]
            logger.debug("Quote: {}", quote)
            return quote
        except Exception as e:
            logger.error("{} Error {}", self.name, e)

    async def get_bid(self, instrument):
        """
        Asynchronously retrieves the bid
        for the specified instrument.

        Args:
            instrument: The instrument for which
            the bid is to be retrieved.

        Returns:
            The bid for the specified instrument.
        """
        try:
            instrument = await self.replace_instrument(instrument)
            return await self.client.fetch_ticker(instrument)["bid"]
        except Exception as e:
            logger.error("{} Error {}", self.name, e)

    async def get_account_balance(self):
        """
        return account balance of
        a given ccxt exchange

        Args:
            None

        Returns:
            balance

        """
        try:
            raw_balance = self.client.fetch_free_balance()
            data = list(raw_balance.items())
            if self.balance_limit:
                data = data[: self.balance_limit_value]
            if filtered_balance := {k: v for k, v in data if v is not None and v > 0}:
                balance_str = "".join(
                    f"{iterator}: {value} \n"
                    for iterator, value in filtered_balance.items()
                )
                return f"{balance_str}"
        except Exception as e:
            logger.error("{} get_account_balance {}", self.name, e)
            return e

    async def get_account_position(self):
        """
        Return account position.
        of a given exchange

        Args:
            None

        Returns:
            position

        """

        try:
            positions = self.client.fetch_positions()
            if positions := [p for p in positions if p["type"] == "open"]:
                return f"{positions}"
        except Exception as e:
            logger.error("{} Error {}", self.name, e)
            return e

    async def pre_order_checks(self, order_params):
        """ """
        return True

    async def get_trading_asset_balance(self):
        """ """
        return self.client.fetchBalance()[f"{self.trading_asset}"]["free"]

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """
        try:
            action = order_params["action"]
            instrument = await self.replace_instrument(order_params["instrument"])
            quantity = order_params.get("quantity", self.trading_risk_amount)
            amount = await self.get_order_amount(
                quantity=quantity,
                instrument=instrument,
                is_percentage=self.trading_risk_percentage,
            )
            params = {
                "stopLoss": {"triggerPrice": order_params.get("stop_loss")},
                "takeProfit": {"triggerPrice": order_params.get("take_profit")},
            }
            if amount and await self.pre_order_checks(order_params):
                order = self.client.create_order(
                    symbol=instrument,
                    type=self.ordertype,
                    side=action,
                    amount=amount,
                    params=params,
                )
                return await self.get_trade_confirmation(order, instrument, action)
            return f"Error executing {self.name}"
        except Exception as e:
            logger.error("{} Error {}", self.name, e)
            return f"Error executing {self.name}"

    # async def execute_order(self, order_params):
    #     """
    #     Execute order

    #     Args:
    #         order_params (dict):
    #             action(str)
    #             instrument(str)
    #             quantity(int)

    #     Returns:
    #         trade_confirmation(dict)

    #     """
    #     try:
    #         action = order_params.get("action")
    #         instrument = await self.replace_instrument(order_params.get("instrument"))
    #         quantity = order_params.get("quantity", self.trading_risk_amount)
    #         logger.debug("quantity {}", quantity)
    #         amount = await self.get_order_amount(
    #             quantity=quantity,
    #             instrument=instrument,
    #             is_percentage=self.trading_risk_percentage,
    #         )
    #         params = {
    #             "stopLoss": {
    #                 "triggerPrice": order_params.get("stop_loss"),
    #                 # "price": order_params.get("action") * 0.98,
    #             },
    #             "takeProfit": {
    #                 "triggerPrice": order_params.get("take_profit"),
    #                 # "price": order_params.get("action") * 0.98,
    #             },
    #         }
    #         logger.debug("amount {}", amount)
    #         pre_order_checks = await self.pre_order_checks(order_params)
    #         logger.debug("pre_order_checks {}", pre_order_checks)

    #         if amount and pre_order_checks:
    #             if order := self.client.create_order(
    #                 symbol=instrument,
    #                 type=self.ordertype,
    #                 side=action,
    #                 amount=amount,
    #                 params=params,
    #             ):
    #                 return await self.get_trade_confirmation(
    # order, instrument, action)
    #         return f"Error executing {self.name}"

    #     except Exception as e:
    #         logger.error("{} Error {}", self.name, e)
    #         return f"Error executing {self.name}"



# """

# etrade client


# """

# import pyetrade

# from ._client import CexClient


# class FxcmHandler(CexClient):
#     """
#     library: https://github.com/jessecooper/pyetrade

#     Args:
#         None

#     Returns:
#         None

#     """

#     def __init__(
#         self,
#         **kwargs,
#     ):
#         """
#         Initialize the client

#         """
#         super().__init__(**kwargs)
#         self.client = pyetrade.ETradeAccounts(
#             self.api_key,
#             self.secret,
#             self.oauth_token,
#             self.oauth_token_secret,
#         )
#         self.account_data = self.client.list_accounts(resp_format="json")
#         self.account_number = self.account_data["Accounts"]["Account"][0]["accountId"]

#     async def get_quote(self, instrument):
#         """
#         Return a quote for a instrument


#         Args:
#             cex
#             instrument

#         Returns:
#             quote
#         """
#         return self.client.get_quote(symbols=[instrument], resp_format="json")

#     async def get_account_balance(self):
#         """
#         return account balance

#         Args:
#             None

#         Returns:
#             balance

#         """

#         return self.client.get_account_balance(account_id_key=self.account_number)

#     async def get_account_position(self):
#         """
#         Return account position.
#         of a given exchange

#         Args:
#             None

#         Returns:
#             position

#         """

#         return self.client.list_transactions(account_id_key=self.account_number)

#     async def get_trading_asset_balance(self):
#         """ """
#         return 0

#     async def execute_order(self, order_params):
#         """
#         Execute order

#         Args:
#             order_params (dict):
#                 action(str)
#                 instrument(str)
#                 quantity(int)

#         Returns:
#             trade_confirmation(dict)

#         """
#         pass



"""

Interactive Brokers client


"""

from ib_insync import IB, IBC, Contract, Order
from loguru import logger

from ._client import CexClient


class IbHandler(CexClient):
    """
    CEX client for IBKR


    Args:
        None

    Returns:
        None

    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initializes the Broker_IBKR_Plugin class.

        This function creates an instance of the IB class from the ib_insync
        library and sets it as the 'ib' attribute of the class.
        It also sets the 'client' attribute to None as a placeholder
        for actual client details.

        To connect to the Interactive Brokers (IBKR) platform,
        the 'connect' method of the 'ib' instance is called.
        This method requires the host, port, and clientId as parameters.
        In this case, the function connects to the IBKR platform using
        the IP address "127.0.0.1", port 7497, and clientId 1.

        After successfully connecting to IBKR, the function logs
        a debug message using the logger module.

        For IBC gateway setup,
        refer to https://github.com/IbcAlpha/IBC/blob/master/userguide.md

        """
        try:
            super().__init__(**kwargs)
            if self.enabled:
                if self.broker_gateway:
                    ibc = IBC(
                        976,
                        gateway=True,
                        tradingMode="paper" if self.testmode else "live",
                        userid=self.user_id,
                        password=self.password,
                    )
                    ibc.start()
                    IB.run()
                self.client = IB()
                self.client.connect(
                    host=self.host,
                    port=self.port,
                    clientId=self.broker_client_id or 1,
                    readonly=False,
                    account=self.broker_account_number or "",
                )
                self.name = self.client.id
                self.account_number = self.client.managedAccounts()[0]
                logger.debug("Connected to IBKR {}", self.client.isConnected())
                logger.debug("Broker_IBKR initialized with account: {}", self.account)

        except Exception as e:
            logger.error("IBC Initialization Error {}", e)
            return None

    async def get_info(self):
        """
        Retrieves information from the accountValues method of the `ib` object.

        Returns:
            The result of calling the accountValues method of the `ib` object.
        """
        return self.client.accountValues()

    async def get_quote(self, instrument):
        """
        Return a quote for a instrument
        of a given ib object


        Args:
            cex
            instrument

        Returns:
            quote
        """
        try:
            instrument = await self.replace_instrument(instrument)

            if contract := self.search_contract(instrument):
                self.client.reqMktData(contract)
                quote = self.client.ticker(contract)
                logger.debug("Quote: {}", quote)
                return quote

        except Exception as e:
            logger.error("{} Error {}", self.name, e)

    async def get_account_balance(self):
        """
        return account balance

        Args:
            None

        Returns:
            balance

        """

        return self.client.accountSummary(self.account)

    async def get_account_position(self):
        """
        Return account position.

        Args:
            None

        Returns:
            position

        """

        try:
            return self.client.positions()
        except Exception as e:
            logger.error("{} Error {}", self.name, e)

    async def pre_order_checks(self, order_params):
        """ """
        return True

    async def get_trading_asset_balance(self):
        """ """
        return self.client.accountSummary(self.account_number)

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """
        try:
            action = order_params.get("action")
            instrument = await self.replace_instrument(order_params.get("instrument"))
            quantity = order_params.get("quantity", self.trading_risk_amount)
            logger.debug("quantity {}", quantity)
            amount = await self.get_order_amount(
                quantity=quantity,
                instrument=instrument,
                is_percentage=self.trading_risk_percentage,
            )

            logger.debug("amount {}", amount)
            pre_order_checks = await self.pre_order_checks(order_params)
            logger.debug("pre_order_checks {}", pre_order_checks)

            if amount and pre_order_checks:
                if contract := self.search_contract(instrument):
                    order = Order()
                    order.action = order_params["action"]
                    order.orderType = order_params["order_type"] or "MKT"
                    order.totalQuantity = amount
                    trade = self.client.placeOrder(contract, order)
                    return await self.get_trade_confirmation(trade, instrument, action)

            return f"Error executing {self.name}"

        except Exception as e:
            logger.error("{} Error {}", self.name, e)
            return f"Error executing {self.name}"

    async def search_contract(self, instrument):
        """
        Asynchronously searches for a contract based on the given instrument.

        Args:
            self: The object instance.
            instrument: The instrument to search for.

        Returns:
            Contract: The contract matching the instrument, or None if not found.
        """
        try:
            if asset := next(
                (
                    item
                    for item in self.mapping
                    if item["id"] == instrument or item["alt"] == instrument
                ),
                None,
            ):
                return Contract(
                    secType=asset["type"],
                    symbol=asset["id"],
                    lastTradeDateOrContractMonth=asset["lastTradeDateOrContractMonth"],
                    strike=asset["strike"],
                    right=asset["right"],
                    multiplier=asset["multiplier"],
                    exchange=asset["exchange"],
                    currency=asset["currency"],
                )
            logger.warning("Asset {} not found in mapping", instrument)
            return None

        except Exception as e:
            logger.error("search_contract {} Error {}", instrument, e)



# """

# Metatrader 5 client


# """

# from loguru import logger
# from mt5linux import MetaTrader5 as mt5

# from ._client import CexClient


# class MetatraderHandler(CexClient):
#     """
#     library: https://github.com/lucas-campagna/mt5linux
#     leveraging metatrader python integration:
#     https://www.mql5.com/en/docs/integration/python_metatrader5/

#     You can use MT5 via the docker image
#     gmag11/metatrader5_vnc which include the python integration
#     if your metatrader container is named metatrader5
#     and the python port is 8001 it will connect

#         Args:
#             None

#         Returns:
#             None

#     """

#     def __init__(
#         self,
#         **kwargs,
#     ):
#         """
#         Initialize the client

#         """
#         super().__init__(**kwargs)
#         self.client = mt5(host=self.host or "metatrader5", port=self.port or 8001)
#         if not self.client.initialize():
#             logger.error("initialize() failed")
#             mt5.shutdown()
#         logger.info(
#             "MT5 initialized\n{}\n{}",
#             self.client.terminal_info(),
#             self.client.version(),
#         )

#     async def get_account_info(self):
#         """
#         Returns:
#             str
#         """
#         self.accounts_data = self.client.get_account_info()
#         return self.accounts_data

#     async def shutdown(self):
#         self.client.shutdown()

#     async def get_account_balance(self):
#         """
#         return account balance

#         Args:
#             None

#         Returns:
#             balance

#         """
#         if account_info := self.get_account_info():
#             return account_info.balance

#     async def get_trading_asset_balance(self):
#         """ """
#         return self.get_account_balance()

#     async def get_account_free_margin(self):
#         """
#         return account balance

#         Args:
#             None

#         Returns:
#             balance

#         """
#         if account_info := self.get_account_info():
#             return account_info.margin_free

#     async def get_account_position(self):
#         """
#         Return account position.
#         of a given exchange

#         Args:
#             None

#         Returns:
#             position

#         """
#         if account_info := self.get_account_info():
#             currency = account_info.currency
#             if positions := self.client.positions():
#                 return [f"{p.symbol} {p.profit} {currency}" for p in positions if p]
#         else:
#             return []

#     async def get_quote(self, instrument):
#         """
#         Return a quote for a instrument


#         Args:
#             cex
#             instrument

#         Returns:
#             quote
#         """
#         if not instrument:
#             raise ValueError("instrument cannot be empty")
#         if quote := self.client.symbol_info_tick(instrument):
#             return quote.bid
#         else:
#             raise ValueError("quote is empty")

#     async def execute_order(self, order_params):
#         """
#         Execute order

#         Args:
#             order_params (dict):
#                 action(str)
#                 instrument(str)
#                 quantity(int)

#         Returns:
#             trade_confirmation(dict)

#         """
#         pass

#     async def calculate_pnl(self, from_date, to_date):
#         """
#         Calculate the PnL for a given period.

#         Parameters:
#             period (str): The period for which
#             to calculate PnL ('W', 'M', 'Y', or None).

#         Returns:
#             pnl: The calculated PnL value.
#         """

#         return self.client.history_orders_get(from_date, to_date)



# """

# Fxcm client


# """


# from loguru import logger

# from ._client import CexClient


# class FxcmHandler(CexClient):
#     """
#     library: https://github.com/fxcm/ForexConnectAPI

#     Args:
#         None

#     Returns:
#         None

#     """

#     def __init__(
#         self,
#         **kwargs,
#     ):
#         """
#         Initialize the client

#         """
#         super().__init__(**kwargs)



#     async def get_quote(self, instrument):
#         """
#         Return a quote for a instrument


#         Args:
#             cex
#             instrument

#         Returns:
#             quote
#         """
#         pass

#     async def get_account_balance(self):
#         """
#         return account balance

#         Args:
#             None

#         Returns:
#             balance

#         """

#         return 0

#     async def get_account_position(self):
#         """
#         Return account position.
#         of a given exchange

#         Args:
#             None

#         Returns:
#             position

#         """


#         return 0

#     async def get_trading_asset_balance(self):
#         """ """
#         return 0

#     async def execute_order(self, order_params):
#         """
#         Execute order

#         Args:
#             order_params (dict):
#                 action(str)
#                 instrument(str)
#                 quantity(int)

#         Returns:
#             trade_confirmation(dict)

#         """
#         pass



"""

CEX client based
class


"""

from datetime import datetime, timedelta

from loguru import logger


class CexClient:
    """
    CEX Object to support CEFI
    exchange and trading platform

    Args:
        None

    Returns:
        None

    Methods:
        get_quote(self, instrument)
        get_offer(self, instrument)
        get_bid(self, instrument)
        get_account_balance(self)
        get_account_position(self)
        calculate_pnl(self, period=None)
        get_account_pnl(self, start_date)
        execute_order(self, order_params)
        get_trading_asset_balance(self)
        get_order_amount(self, quantity, instrument, is_percentage)
        pre_order_checks(self, order_params)
        replace_instrument(self, instrument)
        get_instrument_decimals(self, instrument)
        get_trade_confirmation(self, order_params)


    """

    def __init__(self, **kwargs):
        """
        Initialize the Cex object

        """
        self.protocol = kwargs.get("protocol")
        self.name = kwargs.get("name")
        logger.info("Initializing Client {}", self.name)
        self.enabled = kwargs.get("enabled", False)
        if not self.enabled:
            logger.debug("{} Not enabled", self.name)
            return

        self.__dict__.update(kwargs)
        self.is_pnl_active = kwargs.get("is_pnl_active", False)
        self.client = None
        self.account_number = None
        self.accounts_data = None

    async def get_quote(self, symbol):
        """
        Asynchronously fetches a ask/offer quote
        for the specified instrument.

        :param instrument: The instrument for which the quote is to be fetched.
        :return: The fetched quote.
        """

    # Alias for get_quote
    get_offer = get_quote

    async def get_bid(self, instrument):
        """
        Asynchronously retrieves the bid
        for the specified instrument.

        Args:
            instrument: The instrument for which
            the bid is to be retrieved.

        Returns:
            The bid for the specified instrument.
        """

    async def get_account_balance(self):
        """
        return account balance

        Args:
            None

        Returns:
            balance

        """

    async def get_account_position(self):
        """
        Return account position.

        Args:
            None

        Returns:
            position

        """

    async def get_account_pnl(self, period=None, from_date=None, to_date=None):
        """
        Return account pnl.

        Args:
            None

        Returns:
            pnl
        """
        if from_date is None or to_date is None:
            to_date = datetime.now().date()
            if period is None:
                from_date = to_date
            elif period == "W":
                from_date = to_date - timedelta(days=to_date.weekday())
            elif period == "M":
                from_date = to_date.replace(day=1)
            elif period == "Y":
                from_date = to_date.replace(month=1, day=1)
            else:
                return 0
        try:
            return (
                await self.calculate_pnl(from_date=from_date, to_date=to_date)
                if self.is_pnl_active
                else 0
            )
        except Exception as e:
            logger.error("Error calculating PnL: {}", e)
            return 0

    async def calculate_pnl(self, from_date=None, to_date=None):
        """
        Calculate the PnL for a given start date.

        Parameters:
            from_date: The start date for which to calculate PnL.

        Returns:
            pnl: The calculated PnL value.
        """

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """

    async def get_trading_asset_balance(self):
        """
        Return trading asset balance.

        Args:
            None

        Returns:
            balance

        """

    async def get_order_amount(self, quantity, instrument, is_percentage=True):
        """
        Calculate the order amount based on the risk percentage or money amount.

        Args:
            quantity: The quantity of the order.
            instrument: The instrument of the asset.
            is_percentage: True if quantity is a risk percentage,
            False if it is a money amount.

        Returns:
            The calculated order amount.

        """
        balance = await self.get_trading_asset_balance()
        logger.debug("Balance {}", balance)
        quote = await self.get_quote(instrument)
        logger.debug("Quote {}", quote)

        if not is_percentage and balance and quote:
            logger.debug("Amount based on money {}", balance * quantity / quote)
            return quantity

        if balance and quote:
            risk_percentage = float(quantity) / 100
            amount = balance * risk_percentage / quote
            logger.debug("Amount based on percentage {}", amount)
        if amount >= self.trading_amount_threshold:
            logger.debug("Amount above threshold {}", amount)
            return amount

    async def pre_order_checks(self, order_params):
        """ """

    async def replace_instrument(self, instrument):
        """
        Replace instrument by an alternative instrument, if the
        instrument is not in the mapping, it will be ignored.

        Args:
            order (dict):

        Returns:
            dict
        """
        try:
            for item in self.mapping:
                if item["id"] == instrument:
                    instrument = item["alt"]
                    logger.debug(
                        "Instrument changed from {} to {}", item["id"], instrument
                    )
                    break

        except Exception as e:
            logger.error("{} Error {}", self.name, e)

        return instrument

    async def get_instrument_decimals(self, instrument):
        """
        Get the number of decimal places for the token.

        Returns:
            int: The number of decimal places for the token.

        """

    async def get_trade_confirmation(self, trade, instrument, action):
        """
        Asynchronously retrieves the trade confirmation for a given trade,
        instrument, and action.

        Args:
            self: The object instance
            trade: The trade object
            instrument: The instrument for the trade
            action: The action for the trade

        Returns:
            A string containing the trade confirmation, or None if an error occurs.
        """
        logger.debug("Confirmation {} {} {}", trade, instrument, action)
        try:
            trade_confirmation = (
                f"⬇️ {instrument}" if (action == "SELL") else f"⬆️ {instrument}\n"
            )
            trade_confirmation += f"⚫ {round(0 or trade['amount'], 4)}\n"
            trade_confirmation += f"🔵 {round(0 or trade['price'], 4)}\n"
            trade_confirmation += f"🟢 {round(0 or trade['takeProfitPrice'], 4)}\n"
            trade_confirmation += f"🔴 {round(0 or trade['stopLossPrice'], 4)}\n"
            trade_confirmation += f"ℹ️ {trade['id']}\n"
            trade_confirmation += f"🗓️ {trade['datetime']}"
            if trade_confirmation:
                return f"{trade_confirmation}"
        except Exception as e:
            logger.error("Error {}", e)

    async def modify_position(self, order_params):
        """
        Modify opened position

        Args:
            order_params (dict)

        Returns:
            trade_confirmation(dict)

        """

    async def shutdown(self):
        """ """
        pass



from .capitalcom import CapitalHandler
from .ccxt import CcxtHandler
from .degiro import DegiroHandler

# from .ibkr import IbHandler
from .oanda import OandaHandler

__all__ = [
    "CapitalHandler",
    "CcxtHandler",
    "DegiroHandler",
    # "IbHandler",
    "OandaHandler",
]



"""

Capital.com API client


"""

import asyncio

from capitalcom.client import Client, DirectionType
from capitalcom.client_demo import Client as DemoClient
from loguru import logger

from ._client import CexClient


class CapitalHandler(CexClient):
    """
    Capital.com client
    using
    https://pypi.org/project/capitalcom-python/
    via Capital.com API
    https://open-api.capital.com/
        None

    Returns:
        None

    Methods:
        get_quote(self, instrument)
        get_account_balance(self)
        get_account_position(self)
        pre_order_checks(self, order_params)
        get_trading_asset_balance(self)
        get_instrument_decimals(self, instrument)
        execute_order(self, order_params)


    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initialize the client

        """
        super().__init__(**kwargs)
        self._build_client()

    def _build_client(self):
        """
        Builds and sets the client based on the testmode flag.

        Capital.com session last only 10 minutes

        """
        self.client = DemoClient if self.testmode else Client
        self.client = self.client(
            log=self.user_id,
            pas=self.password,
            api_key=self.api_key,
        )
        logger.debug("Client: {}", self.client)
        if self._fetch_account_data():
            logger.debug("Capital.com client set with account {}", self.account_number)
            return self.client
        else:
            self.client = None
            logger.warning("No capital.com client. Verify settings.")
            return False

    def _fetch_account_data(self):
        """
        Fetches account data from the client,
        extracts the account number,
        and logs account data and session details.

        Parameters:
            self: the CapitalHandler object

        Returns:
            None
        """
        # try:
        #     if self.default_account:
        #         logger.debug("Default account: {}", self.default_account)
        #         self.switch_account(self.default_account)
        #     self.accounts_data = self.client.all_accounts()
        # except Exception as e:
        #     logger.error("Error fetching account data: {}", e)
        #     return False
        # logger.debug("Account data: {}", self.accounts_data)
        # self.account_number = self.accounts_data["accounts"][0]["accountId"]
        # logger.debug("Account number: {}", self.account_number)
        # logger.debug("Session details: {}", self.client.get_sesion_details())
        try:
            self.accounts_data = self.client.all_accounts()
            logger.debug("Account data: {}", self.accounts_data)
            if self.default_account:
                self.switch_account(self.default_account)
                for account in self.accounts_data["accounts"]:
                    if account["accountId"] == self.default_account:
                        self.account_number = account["accountId"]
            else:
                self.account_number = self.accounts_data["accounts"][0]["accountId"]

            logger.debug("Account number: {}", self.account_number)
            logger.debug("Session details: {}", self.client.get_sesion_details())
            return True
        except Exception as e:
            logger.error("Error fetching account data: {}", e)
            return False

    def switch_account(self, account_number):
        """
        Switches to the specified account.

        Parameters:
            account_number: The account number to switch to.

        Returns:
            None
        """
        logger.debug("Switching to account {}", account_number)
        switch = self.client.switch_account(self.default_account)
        logger.debug("Switch: {}", switch)

    async def get_quote(self, instrument):
        """
        Asynchronously fetches a ask/offer quote
        for the specified instrument.

        :param instrument: The instrument for which the quote is to be fetched.
        :return: The fetched quote.
        """
        if self._build_client():
            logger.debug("Instrument: {}", instrument)
            instrument = await self.replace_instrument(instrument)
            logger.debug("Changed Instrument: {}", instrument)
            # search_markets = self.client.searching_market(searchTerm=instrument)
            await asyncio.sleep(1)  # Wait for 1 second
            # logger.debug("Instrument verification: {}", search_markets)

            market = self.client.single_market(instrument)
            # logger.debug("market: {}", market)

            quote = market["snapshot"]["offer"]
            logger.debug("Quote: {}", quote)

            return float(quote)

    # Alias for get_quote
    get_offer = get_quote

    async def get_bid(self, instrument):
        """
        Asynchronously retrieves the bid
        for the specified instrument.

        Args:
            instrument: The instrument for which
            the bid is to be retrieved.

        Returns:
            The bid for the specified instrument.
        """
        if self._build_client():
            logger.debug("Instrument: {}", instrument)
            instrument = await self.replace_instrument(instrument)
            logger.debug("Changed Instrument: {}", instrument)
            # search_markets = self.client.searching_market(searchTerm=instrument)
            await asyncio.sleep(1)  # Wait for 1 second
            # logger.debug("Instrument verification: {}", search_markets)

            market = self.client.single_market(instrument)
            # logger.debug("market: {}", market)

            quote = market["snapshot"]["bid"]
            logger.debug("Quote: {}", quote)

            return float(quote)

    async def get_account_balance(self):
        """
        return account balance of
        a given ccxt exchange

        Args:
            None

        Returns:
            balance

        """
        accounts = self.accounts_data.get("accounts", [])

        balances = [
            f"{account['accountName']}: {account['balance']['balance']}\n"
            for account in accounts
            if "balance" in account and account["balance"].get("balance") is not None
        ]

        return "".join(balances)

    async def get_account_position(self):
        """
        Return account position.
        of a given exchange

        Args:
            None

        Returns:
            position

        """
        if positions := self.client.all_positions():
            extracted_positions = []
            for position_data in positions.get("positions", []):
                position_details = position_data.get("position", {})
                market_details = position_data.get("market", {})
                epic = market_details.get("epic", "")
                upl = position_details.get("upl", 0)
                extracted_positions.append(f"{epic}: {upl}")
            return "\n".join(extracted_positions)

    async def calculate_pnl(self, period=None):
        """
        no pnl info available via openapi endpoint

        Args:
            None

        Returns:
            pnl
        """
        # end_date = datetime.now()
        # formatted_end_date = end_date.strftime("%Y-%m-%dT%H:%M:%S")
        # logger.debug("{} {}", start_date, formatted_end_date)
        # history = self.client.account_activity_history(
        #     fr=start_date, to=formatted_end_date, detailed=True
        # )
        # logger.debug("History: {}", history)

        return 0

    async def pre_order_checks(self, order_params):
        """ """
        return True

    async def get_trading_asset_balance(self):
        """
        Return the available balance of the trading asset from the account.

        Returns:
            float: The available balance of the trading asset.
        """
        return next(
            (
                account["balance"]["available"]
                for account in self.accounts_data["accounts"]
                if account["accountName"] == self.trading_asset
            ),
            0,
        )

    async def get_instrument_decimals(self, instrument):
        """
        Get the number of decimal places for the instrument.

        Returns:
            int: The number of decimal places for the instrument.
        """
        instrument_info = self.client.single_market(instrument)
        decimals = instrument_info.get("snapshot", {}).get("decimalPlacesFactor", 0)
        logger.debug("Decimals {}", decimals)
        return int(decimals)

    async def get_instrument_min_amount(self, instrument):
        """
        Get the minimum amount needed for a specific instrument.

        Args:
            instrument (str): The instrument for which the minimum amount is needed.

        Returns:
            float: The minimum amount needed for the specified instrument.
        """
        instrument_info = self.client.single_market(instrument)
        logger.debug("instrument_info {}", instrument_info)
        minimum_amount = (
            instrument_info.get("dealingRules", {})
            .get("minDealSize", {})
            .get("value", 0)
        )
        logger.debug("Minimum Amount Needed {}", minimum_amount)
        return float(minimum_amount)

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """
        try:
            action_str = order_params.get("action")
            action = DirectionType[action_str]
            instrument = await self.replace_instrument(order_params.get("instrument"))

            quantity = order_params.get("quantity", self.trading_risk_amount)
            amount = await self.get_order_amount(
                quantity=quantity,
                instrument=instrument,
                is_percentage=self.trading_risk_percentage,
            )
            min_amount = await self.get_instrument_min_amount(instrument)
            amount = max(amount, min_amount)
            logger.debug("Amount to execute {}", amount)
            await asyncio.sleep(1)  # Wait for 1 second

            if not (await self.pre_order_checks(order_params)):
                return f"Error executing {self.name}"

            decimals = await self.get_instrument_decimals(instrument)
            await asyncio.sleep(1)  # Wait for 1 second

            offer = await self.get_offer(instrument)
            await asyncio.sleep(1)  # Wait for 1 second

            bid = await self.get_bid(instrument)
            await asyncio.sleep(1)  # Wait for 1 second

            logger.debug("bid {}", bid)
            logger.debug("offer {}", offer)
            profit_price = (
                (offer + (int(order_params.get("take_profit", 0)) / (10**decimals)))
                if action_str == "BUY"
                else (bid - (int(order_params.get("take_profit", 0)) / (10**decimals)))
            )
            stop_price = (
                (bid - (int(order_params.get("stop_loss", 0)) / (10**decimals)))
                if action_str == "BUY"
                else (offer + (int(order_params.get("stop_loss", 0)) / (10**decimals)))
            )
            logger.debug("stop price {}", stop_price)
            logger.debug("profit price {}", profit_price)
            order = self.client.place_the_position(
                direction=action,
                epic=instrument,
                size=amount,
                gsl=False,
                tsl=False,
                stop_level=stop_price,
                stop_distance=None,
                stop_amount=None,
                profit_level=profit_price,
                profit_distance=None,
                profit_amount=None,
            )

            if "errorCode" in order:
                logger.error(f"Error placing order: {order['errorCode']}")
                return str(order["errorCode"])

            logger.debug("Order: {}", order)
            deal_reference = order["dealReference"]
            order_check = self.client.position_order_confirmation(
                deal_reference=deal_reference
            )

            trade = {
                "amount": order_check.get("size", 0),
                "price": order_check.get("level", 0),
                "takeProfitPrice": profit_price,
                "stopLossPrice": stop_price,
                "id": order_check.get("dealId", ""),
                "datetime": order_check.get("date", ""),
            }
            return await self.get_trade_confirmation(trade, instrument, action)

        except Exception as e:
            logger.error("{} Error {}", self.name, e)
            return f"Error executing {self.name}: {e}"

    async def modify_position(self, order_params):
        """
        Modify parameters such as SL / TP of a position that is opened
        No capability to modify amount to reduce

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)
                stop_price(int)
                stop_distance(int)
                stop_amount(int)
                take_profit_price(int)
                take_profit_distance(int)
                take_profit_amount(int)

        Returns:
            trade_confirmation(dict)

        """
        try:
            order = self.client.update_the_position(
                dealid=order_params["id"],
                gsl=order_params.get("trailing_stop_enabled", False),
                tsl=order_params.get("trailing_stop_enabled", False),
                stop_level=order_params.get("stop_price"),
                stop_distance=order_params.get("stop_distance"),
                stop_amount=order_params.get("stop_amount"),
                profit_level=order_params.get("take_profit_price"),
                profit_distance=order_params.get("take_profit_distance"),
                profit_amount=order_params.get("take_profit_amount"),
            )

            if "errorCode" in order:
                logger.error(f"Error modifying order: {order['errorCode']}")
                return str(order["errorCode"])

            logger.debug("Order: {}", order)
            deal_reference = order["dealReference"]
            order_check = self.client.position_order_confirmation(
                deal_reference=deal_reference
            )

            trade = {
                "amount": order_check.get("size", 0),
                "price": order_check.get("level", 0),
                "id": order_check.get("dealId", ""),
                "datetime": order_check.get("date", ""),
            }
            return await self.get_trade_confirmation(
                trade, order_params.get("instrument"), order_params.get("action")
            )
        except Exception as e:
            logger.error("{} Error {}", self.name, e)
            return f"Error executing {self.name}: {e}"



"""

Degiro client


"""

from datetime import date

from degiro_connector.trading.api import API as TradingAPI
from degiro_connector.trading.models.account import Format, ReportRequest
from degiro_connector.trading.models.credentials import build_credentials
from degiro_connector.trading.models.order import Action, Order, OrderType, TimeType
from loguru import logger

from ._client import CexClient


class DegiroHandler(CexClient):
    """
    CEX client
    via Degiro API
    https://github.com/Chavithra/degiro-connector

    Args:
        None

    Returns:
        None

    """

    def __init__(
        self,
        **kwargs,
    ):
        """
        Initialize the client

        """
        super().__init__(**kwargs)
        if self.name is None:
            return
        credentials = build_credentials(
            # location="config/config.json",
            override={
                "username": self.user_id,
                "password": self.password,
                "int_account": self.broker_account_number,  # From `get_client_details`
                "totp_secret_key": self.secret,  # For 2FA
            },
        )
        client = TradingAPI(credentials=credentials)
        client.connect()
        self.accounts_data = client.get_account_info()

    async def get_quote(self, instrument):
        """
        Asynchronously fetches a ask/offer quote
        for the specified instrument.

        :param instrument: The instrument for which the quote is to be fetched.
        :return: The fetched quote.
        """
        # TODO
        # try:
        #     instrument = await self.replace_instrument(instrument)

        #     ticker = self.client.
        #     quote =
        #     logger.debug("Quote: {}", quote)
        #     return quote
        # except Exception as e:
        #     logger.error("{} Error {}", self.name, e)

    async def get_account_balance(self):
        """
        return account balance

        Args:
            None

        Returns:
            balance

        """
        return self.client.get_account_report(
            report_request=ReportRequest(
                country="FR",
                lang="fr",
                format=Format.CSV,
                from_date=date(year=date.today().year - 1, month=1, day=1),
                to_date=date.today(),
            ),
            raw=False,
        )

    async def get_account_position(self):
        """
        Return account position.

        Args:
            None

        Returns:
            position

        """
        return self.client.get_position_report(
            report_request=ReportRequest(
                country="FR",
                lang="fr",
                format=Format.XLS,
                from_date=date(year=date.today().year - 1, month=1, day=1),
                to_date=date.today(),
            ),
            raw=False,
        )

    async def pre_order_checks(self, order_params):
        """ """
        return True

    async def get_trading_asset_balance(self):
        """ """
        # TODO
        # return self.client.fetchBalance()[f"{self.trading_asset}"]["free"]

    async def execute_order(self, order_params):
        """
        Execute order

        Args:
            order_params (dict):
                action(str)
                instrument(str)
                quantity(int)

        Returns:
            trade_confirmation(dict)

        """
        # TODO: Test and fix order call
        try:
            action = order_params.get("action")
            instrument = await self.replace_instrument(order_params.get("instrument"))
            quantity = order_params.get("quantity", self.trading_risk_amount)
            if not action or not instrument:
                raise ValueError("Missing required parameters: action or instrument")
            logger.debug(f"quantity {quantity}")
            amount = await self.get_order_amount(
                quantity=quantity,
                instrument=instrument,
                is_percentage=self.trading_risk_percentage,
            )
            if not amount:
                raise ValueError("Failed to calculate order amount")
            logger.debug(f"amount {amount}")
            pre_order_checks = await self.pre_order_checks(order_params)
            if not pre_order_checks:
                raise ValueError("Pre-order checks failed")
            order = Order(
                buy_sell=Action.BUY if action == "BUY" else Action.SELL,
                order_type=OrderType.LIMIT,
                price=12.1,
                product_id=72160,
                size=amount,
                time_type=TimeType.GOOD_TILL_DAY,
            )
            if checking_response := self.client.check_order(order=order):
                if confirmation_response := self.client.confirm_order(
                    confirmation_id=checking_response.confirmation_id,
                    order=order,
                ):
                    logger.debug("Confirmation: {}", confirmation_response)
                    return await self.get_trade_confirmation(order, instrument, action)
        except Exception as e:
            logger.error(f"{self.name} Error {e}")
            return f"Error executing {self.name}"
